class SourceWrap:
    r"""Helper class to relax requirements for a source.
    It solves two problems:
        1) the source doesn't have to supply the index key in data
        2) __getitem__ provided for sources that only contain __iter__.
            It's fast when accesing elements in increasing order.
    """

    def __init__(self, source_instance):
        self.instance = source_instance
        self._hack_iter = None
        self._hack_next_idx = None

    def __getitem__(self, idx):
        if hasattr(self.instance, "__getitem__"):
            return {**self.instance[idx], "idx": idx}
        return self.hack_getitem(idx)

    def __iter__(self):
        for idx, data in enumerate(self.instance):
            yield {**data, "idx": idx}

    def hack_getitem(self, idx):

        if self._hack_iter is None or self._hack_next_idx > idx:
            self._hack_iter = iter(self.instance)
            self._hack_next_idx = 0
        diff = idx - self._hack_next_idx
        for _ in range(diff):
            self._hack_next_idx += 1
            _ = next(self._hack_iter)
        self._hack_next_idx += 1
        return {**next(self._hack_iter), "idx": idx}



from dame.source import SourceWrap


class MockSource:
    def __iter__(self):
        for i in range(10):
            yield {"what": i + 10}


def test_hack_getitem():
    src = SourceWrap(MockSource())
    for _ in range(2):
        for j in range(10):
            assert src[j] == {"what": j + 10, "idx": j}



from dame.versionable import Versionable


class PlusOne(Versionable):
    provides = ("p1",)

    def apply(self, *, number):
        return {"p1": number + 1}


class PlusTwo(Versionable):
    provides = ("p2",)

    def apply(self, *, p1):
        return {"p2": p1 + 1}


class PlusXN(Versionable):

    provides = ("pxn",)

    def __init__(self, x, n=0):
        self.x = x
        self.n = n

    def apply(self, *, number):
        return {"pxn": number + self.x ** self.n}


class ThreeNums:
    provides = ("number",)

    def __getitem__(self, idx):
        if idx > 2 or idx < 0:
            raise IndexError
        return {"number": idx}

    def __len__(self):
        return 3



from contextlib import contextmanager
from tempfile import TemporaryDirectory
from io import BytesIO
import pickle

import numpy as np

from dame.storage import PeeWeeStore, NumpyPlaceholder

from .test_classes import PlusXN


@contextmanager
def tmp_db_path():
    with TemporaryDirectory() as tmpdir:
        yield f"{tmpdir}/db.sqlite3"


def test_separate_numpy_data():
    r1, r2 = np.random.rand(2, 10)
    data = {"a": 1, "na": r1, "b": 2, "nb": r2}
    newdata, np_data = PeeWeeStore.separate_numpy_data(data)
    assert newdata["a"] == 1 and newdata["b"] == 2
    assert (np_data == [r1, r2]) or (np_data == [r2, r1])
    assert isinstance(newdata["na"], NumpyPlaceholder) and isinstance(
        newdata["nb"], NumpyPlaceholder
    )
    assert np.all(np_data[newdata["na"].num] == r1)
    assert np.all(np_data[newdata["nb"].num] == r2)


def test_unpack_blobs():
    data = {"a": 1, "b": 2, "c": NumpyPlaceholder(0), "d": NumpyPlaceholder(1)}
    data_blob = pickle.dumps(data)
    np_blob = BytesIO()
    arr1, arr2 = np.random.rand(2, 10)
    np.save(np_blob, arr1)
    np.save(np_blob, arr2)
    np_blob = np_blob.getvalue()
    unpacked_data = PeeWeeStore.unpack_blobs(data_blob, np_blob)
    assert np.all(unpacked_data["c"] == arr1)
    assert np.all(unpacked_data["d"] == arr2)
    del unpacked_data["c"]
    del unpacked_data["d"]
    del data["c"]
    del data["d"]
    assert data == unpacked_data


def test_storage():
    with tmp_db_path() as db_path:
        transform = PlusXN(3)
        store = PeeWeeStore()
        store.open(db_path, pragmas={"foreign_keys": 1})
        data = {"random": np.random.rand(5), "foo": "bar"}
        store.save(0, transform, data)
        recv = store.load(0, transform)
        store.close()
        assert np.all(recv["random"] == data["random"])
        del recv["random"]
        del data["random"]
        assert recv == data



from inspect import getsource
from hashlib import sha256
import secrets
import string


class Versionable:
    """TODO write documentation."""

    def register_params(self, **kwargs):
        if getattr(self, "_versionables", None) is None:
            self._versionables = {}
        self._versionables.update(**kwargs)

    def version(self):
        """Detects changes in subclass's code and parameters.

        Dame automagically invalidates cache and storage on changes in Transforms.
        For this to work Dame needs a way to determine whether the Transform had
        changed since the last use.

        Returns:
            str: Hash computed over source code of the class and it's parameters
        """
        if hasattr(self, "__version_str__") and self.__version_str__ is not None:
            return self.__version_str__
        result = sha256()
        # TODO strip docs and comments from getsource. Maybe using astunparse?
        result.update(getsource(self.__class__).encode())
        result.update(self.get_params_hash().encode())
        return result.hexdigest()

    def get_params_hash(self):
        result = sha256()
        for key, value in getattr(self, "_versionables", {}).items():
            result.update(key.encode())
            result.update(self.get_version_str(value).encode())
        return result.hexdigest()

    def get_version_str(self, x):
        if hasattr(x, "__version_str__"):
            return x.__version_str__
        elif hasattr(x, "version"):
            return x.version()
        # TODO issue a warning for objects (str(x) == "<some_class object at 0x838>")
        return str(x)


class Unversionable(Versionable):
    """TODO write documentation."""

    @property
    def __version_str__(self):
        if not hasattr(self, "__version_str_gen__"):
            # https://stackoverflow.com/questions/2257441/random-string-generation-with-upper-case-letters-and-digits/23728630#23728630
            random_str = "".join(
                secrets.choice(string.ascii_uppercase + string.digits)
                for _ in range(100)
            )
            res = sha256()
            res.update(random_str.encode())
            self.__version_str_gen__ = res.hexdigest()
        return self.__version_str_gen__



from datetime import datetime
from io import BytesIO
import pickle

import numpy as np
from peewee import (
    Model,
    FixedCharField,
    ForeignKeyField,
    DateTimeField,
    BlobField,
    IntegerField,
    SqliteDatabase,
)


class TransformModel(Model):
    name = FixedCharField(max_length=128)
    digest = FixedCharField(max_length=128)
    version = DateTimeField(default=datetime.now)


class Result(Model):
    origin = ForeignKeyField(TransformModel, backref="results")
    dataset_index = IntegerField()
    pickled_data = BlobField()
    numpy_data = BlobField(null=True)


class NumpyPlaceholder:
    def __init__(self, num):
        self.num = num


class PeeWeeStore:
    def __init__(
        self, db_cls=SqliteDatabase, src_id=None, db_kwargs=None, db_name=None
    ):
        self.db_cls = db_cls
        self.db_name = db_name if db_name is not None else src_id
        self.db_kwargs = db_kwargs or dict({"pragmas": {"foreign_keys": 1}})

    def open(self, db_name=None, **kwargs):
        if db_name is None:
            db_name = self.db_name
        if not kwargs:
            kwargs = self.db_kwargs
        self.db = self.db_cls(db_name, **kwargs)
        # Models in relation are automatically bound
        TransformModel.bind(self.db)
        self.db.connect()
        self.db.create_tables([TransformModel, Result])

    def close(self):
        self.db.close()

    def transform_id(self, t):
        t_name = t.__class__.__name__
        if not hasattr(self, "_transform_ids"):
            self._transform_ids = {}
        if t_name not in self._transform_ids:
            self._transform_ids[t_name] = TransformModel.get_or_create(
                digest=t.version(), name=t_name
            )[0].id
        return self._transform_ids[t_name]

    @staticmethod
    def separate_numpy_data(data):
        np_data = []
        for key in sorted(data.keys()):
            if isinstance(data[key], np.ndarray):
                np_data.append(data[key])
                data[key] = NumpyPlaceholder(len(np_data) - 1)
        return data, np_data

    @staticmethod
    def get_blobs(data, np_data):
        out = pickle.dumps(data)
        if len(np_data) > 0:
            np_out = BytesIO()
            for arr in np_data:
                np.save(np_out, arr)
            return out, np_out.getvalue()
        return out, None

    @staticmethod
    def unpack_blobs(data_blob, np_data_blob):
        data = pickle.loads(data_blob)
        if np_data_blob is not None:
            ord_keys = [i for i in data.items() if isinstance(i[1], NumpyPlaceholder)]
            ord_keys = sorted(ord_keys, key=lambda item: item[1].num)
            ord_keys, _ = zip(*ord_keys)
            np_data_blob = BytesIO(np_data_blob)
            for key in ord_keys:
                data[key] = np.load(np_data_blob)
        return data

    def save(self, idx, transform, data):
        data = dict(data)
        out, np_out = self.get_blobs(*self.separate_numpy_data(data))
        trans = self.transform_id(transform)
        res = Result(
            origin_id=trans, dataset_index=idx, pickled_data=out, numpy_data=np_out,
        )
        res.save()

    def load(self, idx, transform):
        trans = self.transform_id(transform)
        res = Result.get_or_none(origin_id=trans, dataset_index=idx)
        if res is None:
            return None
        return self.unpack_blobs(res.pickled_data, res.numpy_data)



from inspect import getfullargspec


class Stages:
    """DAG functionality for Dataset's transforms."""

    def __init__(self, source, transforms):
        self.stages = self.topsort(source, list(transforms))
        self.source = source

    def get_requirements(self, transform):
        return getfullargspec(transform.apply)[4]

    def topsort(self, source, transforms):
        """Sorts the transforms topologistagcally."""
        provider = {key: t for t in transforms for key in t.provides}
        self.provider = provider

        dependants = {t: set() for t in transforms}
        for t in transforms:
            for key in self.get_requirements(t):
                if key not in getattr(source, "provides", tuple()):
                    dependants[provider[key]].add(t)

        requires = {
            t: set(self.get_requirements(t)) - set(source.provides) for t in transforms
        }
        Q = [t for t, deps in requires.items() if len(deps) == 0]
        ordered = []
        while Q:
            t = Q.pop()
            ordered.append(t)
            for dependant in dependants[t]:
                provides = set(t.provides)
                requires[dependant] -= provides
                if len(requires[dependant]) == 0:
                    Q.append(dependant)
        assert len(ordered) == len(transforms), "Something is wrong with topsort!"
        return ordered

    def __iter__(self):
        return iter(self.stages)

    def __getitem__(self, idx):
        return self.stages[idx]

    def to(self, *keywords):
        result = set()
        Q = list([self.provider[kw] for kw in keywords])
        while Q:
            t = Q.pop()
            if t not in result:
                result.add(t)
                Q.extend(
                    set(
                        self.provider[kw]
                        for kw in self.get_requirements(t)
                        if kw not in getattr(self.source, "provides", tuple())
                    )
                )
        return filter(lambda t: t in result, self.stages)



from dame.stages import Stages

from .test_classes import PlusOne, PlusTwo, ThreeNums


def test_dag():
    stages = Stages(ThreeNums, (PlusTwo, PlusOne))
    assert list(iter(stages)) == [
        PlusOne,
        PlusTwo,
    ]
    assert list(stages.to("p1")) == [PlusOne]



from multiprocessing import Pool
from multiprocessing.util import Finalize

from .stages import Stages
from .source import SourceWrap


def make_instance_with_context(cls, context):
    if cls.__name__ in context:
        ctx = context[cls.__name__]
        return cls(*ctx.get("args", []), **ctx.get("kwargs", {}))
    return cls()


class SequentialWorker:
    r"""Performs all the necessary computations in a current process"""

    @staticmethod
    def make_instance(stages, context, store):
        SequentialWorker.instance = SequentialWorker(
            stages, context, global_store=store
        )
        SequentialWorker.instance.__enter__()
        # https://stackoverflow.com/a/24724452
        Finalize(
            SequentialWorker.instance,
            SequentialWorker.instance.__exit__,
            args=(None, None, None),
            exitpriority=10,
        )

    def __init__(self, stages, context, global_store=None):
        self.stages = stages
        self.context = context
        if global_store is not None:
            self.store = make_instance_with_context(global_store, context)

    def __enter__(self):
        if hasattr(self, "store"):
            self.store.open()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if hasattr(self, "store"):
            self.store.close()

    @property
    def stage_instances(self):
        if not hasattr(self, "_stage_instances"):
            self._stage_instances = [
                make_instance_with_context(stage_cls, self.context)
                for stage_cls in self.stages
            ]
        return self._stage_instances

    @staticmethod
    def instance_compute_to(d, kw):
        return SequentialWorker.instance.compute_to(d, kw)

    def compute_to(self, data, keyword):
        r"""Computes an element upto the transform T that provides the keyword (including T).

        Args:
            data (dict): data from source
            keyword (string): Computation stops while the data has keyword
        
        Returns:
            dict: data after transformation via all declared transforms
        """
        cls_it = self.stages.to(keyword)
        stage_it = iter(self.stage_instances)
        while keyword not in data:
            stage = next(stage_it)
            next_cls = next(cls_it)
            while not isinstance(stage, next_cls):
                stage = next(stage_it)
            data = self.compute_stage(data, stage)
        return data

    @staticmethod
    def instance_compute_full(d):
        return SequentialWorker.instance.compute_full(d)

    def compute_full(self, data):
        r"""Computes an element using all the transforms
        
        Args:
            data (dict): data from source
        
        Returns:
            dict: data after transformation via all declared transforms
        """
        for stage in self.stage_instances:
            data = self.compute_stage(data, stage)
        return data

    @staticmethod
    def instance_compute_stage(d, s):
        return SequentialWorker.instance.compute_stage(d, s)

    def load_from_store(self, idx, transform):
        if not hasattr(self, "store"):
            return None
        return self.store.load(idx, transform)

    def save_to_store(self, idx, transform, data):
        if not hasattr(self, "store"):
            return
        self.store.save(idx, transform, data)

    def compute_stage(self, data, stage):
        r"""Computes a single stage
        
        Args:
            data (dict): data from previous transforms
            stage (Transform): Transformation to apply
        
        Returns:
            dict: updated data
        """
        new_data = self.load_from_store(data["idx"], stage)
        if new_data is not None:
            data.update(new_data)
            return data

        kwargs = {
            key: value
            for key, value in data.items()
            if key in self.stages.get_requirements(stage)
        }
        new_data = stage.apply(**kwargs)
        data.update(new_data)
        self.save_to_store(data["idx"], stage, new_data)
        return data


class PoolJoiningIterator:
    def __init__(self, base_iter, pool):
        self.it = base_iter
        self.pool = pool

    def __iter__(self):
        try:
            yield from self.it
        finally:
            self.pool.close()
            self.pool.join()


class WorkManager:
    """Manages the spawning of workers in separate processes."""

    def __init__(self, source, transforms, context, n_processes=None, store=None):
        self.n_processes = n_processes
        self.source_instance = source
        self.context = context
        self.stages = Stages(source, transforms)
        self.store = store

    @property
    def wrapped_source_instance(self):
        if not hasattr(self, "_wrapped_source_instance"):
            self._wrapped_source_instance = SourceWrap(self.source_instance)
        return self._wrapped_source_instance

    def fast_compute(self):
        pool = Pool(
            processes=self.n_processes,
            initializer=SequentialWorker.make_instance,
            initargs=(self.stages, self.context, self.store),
        )

        res_it = pool.imap(
            SequentialWorker.instance_compute_full,
            (data for data in self.wrapped_source_instance),
        )

        return iter(PoolJoiningIterator(res_it, pool))

    def compute_one(self, idx):
        with SequentialWorker(self.stages, self.context, global_store=self.store) as w:
            return w.compute_full(self.wrapped_source_instance[idx])



# TODO show some love to these tests, rethink, cleanup and write the docs!
from tempfile import TemporaryDirectory
from pathlib import Path
from importlib import import_module, reload
import fileinput


from dame.versionable import Versionable, Unversionable


code_base = """
from dame.versionable import Versionable

class TestT(Versionable):
    {}
{}
    def __init__(self, **kwargs):
        self.register_params(**kwargs)
        for key, value in kwargs.items():
            setattr(self, key, value)
{}
"""


dir_path = Path(__file__).resolve().parent
tmp_dir = TemporaryDirectory(dir=dir_path)
open(Path(tmp_dir.name) / "__init__.py", "w").close()


def make_versionable(doc, attrs, apply):
    def indent(code, spaces):
        return "".join(f"{' ' * spaces}{line}\n" for line in code.split("\n"))

    doc_code = f'"""{doc}"""'
    attr_code = "".join(f"{key} = {value}\n" for key, value in attrs.items())
    attr_code = indent(attr_code, 4)
    apply_code = indent(apply, 8)[4:]
    return code_base.format(doc_code, attr_code, apply_code)


class tmp_module:
    def __init__(self, name, content):
        self.fname = f"{name}.py"
        self.mod_name = Path(tmp_dir.name).name
        self.path = Path(tmp_dir.name) / self.fname
        self.content = content

    def __enter__(self):
        self.f = open(self.path, "w+")
        self.f.write(self.content)
        self.f.close()
        imp_name = f".{self.mod_name}.{self.fname[:-3]}"
        self.mod = import_module(imp_name, package="tests")
        return self.mod

    def __exit__(self, exc_type, exc_value, traceback):
        return False


def test_basic():
    with tmp_module("transformA", make_versionable("", {}, "")) as ta:
        tae1 = ta.TestT()
        tae2 = ta.TestT()

        assert tae1.version() == tae2.version()


def test_params():
    with tmp_module("transformA", make_versionable("", {}, "")) as ta:

        tak1 = ta.TestT(ala="kota")
        tak2 = ta.TestT(basia="kota")
        tak3 = ta.TestT(ala="psa")
        tak4 = ta.TestT(ala="kota")

        assert tak1.version() == tak4.version()
        assert tak1.version() != tak2.version()
        assert tak1.version() != tak3.version()
        assert tak2.version() != tak3.version()


def test_code():
    code = "def whatever():\npass"

    with tmp_module("transA", make_versionable("", {}, code)) as ta:
        tt = ta.TestT()
        ver1 = tt.version()
        for line in fileinput.input(files=(ta.__file__,), inplace=True):
            if "whatever" in line:
                print("    def whatever2():")
            else:
                print(line[:-1])
        ta = reload(ta)
        tt2 = ta.TestT()
        ver2 = tt2.version()

        assert ver1 != ver2


class X(Versionable):
    def x():
        pass


class Y(Versionable):
    def y():
        pass


def test_version_str():
    x, y = X(), Y()
    assert x.version() != y.version()
    x.__version_str__ = "1"
    y.__version_str__ = "1"
    assert x.version() == y.version()


class Z(Unversionable):
    pass


def test_unversionable():
    p, q = Z(), Z()
    assert p.version() != q.version()



from pytest import raises

from dame import Dataset

from .test_classes import ThreeNums, PlusOne, PlusTwo, PlusXN


class StandardDataset(Dataset):
    source = ThreeNums()
    transforms = (PlusOne, PlusTwo)


def expected(num):
    return {"number": num, "p1": num + 1, "p2": num + 2, "idx": num}


def test_gets_item():
    data = StandardDataset()
    assert data[0] == expected(0)
    assert data[2] == expected(2)


def test_len():
    assert len(StandardDataset()) == 3


full = [expected(x) for x in range(3)]


def test_iter():
    assert list(StandardDataset()) == full


def test_assequence():
    data = StandardDataset()
    assert list(data.assequence()) == full
    assert list(data.assequence(of="p1")) == [{"p1": d["p1"]} for d in full]
    assert list(data.assequence(of=set(["p1", "p2"]))) == [
        {"p1": d["p1"], "p2": d["p2"]} for d in full
    ]


def test_safety_no_source():
    with raises(AssertionError):

        class NoSource(Dataset):
            pass

        for i in NoSource():
            pass


def test_safety_overlapping_keywords():
    with raises(AssertionError):

        class RepeatT(Dataset):
            transforms = (PlusOne, PlusOne)

        for i in RepeatT():
            pass


class OtherDataset(StandardDataset):
    transforms = (PlusOne, PlusXN)


def test_right_args():
    data = OtherDataset()
    data.set_arguments_for(PlusXN, 2, n=10)
    assert data[0] == {"number": 0, "p1": 1, "pxn": 1024, "idx": 0}
    data.set_arguments_for(PlusXN, 2, n=9)
    assert data[0] == {"number": 0, "p1": 1, "pxn": 512, "idx": 0}



from dame.worker import SequentialWorker, WorkManager
from dame.stages import Stages
from dame.source import SourceWrap
from .test_classes import ThreeNums, PlusOne, PlusTwo, PlusXN


class MockStages(Stages):
    def __init__(self):
        pass

    def to(self, keyword):
        if keyword == "p1":
            return iter([PlusOne])
        if keyword == "pxn":
            return iter([PlusXN])
        return iter([PlusOne, PlusTwo])

    def __iter__(self):
        yield PlusOne
        yield PlusXN
        yield PlusTwo


class MockStorage:
    def __init__(self, *args, **kwargs):
        pass

    def open(self):
        self.data = {}
        print("Opening")

    def save(self, idx, transform, data):
        print(f"Saving {idx}.{transform.__class__.__name__}")
        if idx not in self.data:
            self.data[idx] = {}
        self.data[idx][transform.__class__.__name__] = data

    def load(self, idx, transform):
        print(f"Loading {idx}.{transform.__class__.__name__}")
        if idx == 0:
            if transform.__class__ == PlusOne:
                print("Returning")
                return {"p1": 1}
            elif transform.__class__ == PlusTwo:
                print("Returning")
                return {"p2": 2}
        return self.data.get(idx, {}).get(transform.__class__.__name__, None)

    def close(self):
        print(f"Finalizing")
        delattr(self, "data")


def test_computations():
    stages = MockStages()
    source = SourceWrap(ThreeNums())
    context = {PlusXN.__name__: {"args": [10], "kwargs": {"n": 3}}}
    worker = SequentialWorker(stages, context)

    data_1 = {**source[0], "p1": 1}
    data_2 = {**data_1, "p2": 2}
    data_full = {**data_2, "pxn": 10 ** 3}

    assert worker.compute_to(source[0], "p1") == data_1
    assert worker.compute_to(source[0], "p2") == data_2

    assert worker.compute_full(source[0]) == data_full

    assert worker.compute_stage(dict(data_1), PlusTwo()) == data_2
    assert worker.compute_stage(source[0], PlusOne()) == data_1


def test_uses_store(capfd):
    source = ThreeNums()
    context = {PlusXN.__name__: {"args": [10]}}
    transforms = (PlusOne, PlusXN, PlusTwo)
    manager = WorkManager(source, transforms, context, n_processes=1, store=MockStorage)

    for i in manager.fast_compute():
        pass

    out = capfd.readouterr()
    expected_out = "Opening\nLoading 0.PlusXN\nSaving 0.PlusXN\nLoading 0.PlusOne\nReturning\nLoading 0.PlusTwo\nReturning\nLoading 1.PlusXN\nSaving 1.PlusXN\nLoading 1.PlusOne\nSaving 1.PlusOne\nLoading 1.PlusTwo\nSaving 1.PlusTwo\nLoading 2.PlusXN\nSaving 2.PlusXN\nLoading 2.PlusOne\nSaving 2.PlusOne\nLoading 2.PlusTwo\nSaving 2.PlusTwo\nFinalizing\n"
    assert out.out == expected_out



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="dame",
    version="0.0.2",
    author="Stanisław Morawski",
    author_email="stas.morawski@gmail.com",
    description="Manage your dataflows seamlessly",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/malpunek/dame",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 1 - Planning",
        "Topic :: Software Development"
    ],
    python_requires='>=3.6',
    install_requires=['numpy', 'peewee>=3'],
)



from itertools import chain
from collections import Counter

from .worker import WorkManager


class Dataset:
    r"""The most important class in DAME.

    Manages access, computation of dataset elements

    What you probably want to do is to subclass this class and redefine it's source
    and transforms attributes.

    Attributes:
        source (Source): As lightweight as possible data source
        transforms (Iterable[Transform_cls]): Processing of data items at element level
        store (Store_cls): Storage method to permanently save computed

    """
    source = None
    transforms = tuple()
    context = {}

    def __init__(self):
        if hasattr(self, "store"):
            self.add_arguments_for(self.store, src_id=self.source.version())

    @property
    def manager(self):
        if getattr(self, "_manager", None) is None:
            if not hasattr(self, "_validated"):
                self.validate()
                self._validated = True
            self._manager = WorkManager(
                self.source,
                self.transforms,
                self.context,
                n_processes=getattr(self, "n_processes", None),
                store=getattr(self, "store", None)
            )
        return self._manager

    def __getitem__(self, idx):
        r"""Computes a single dataset element"""
        return self.manager.compute_one(idx)

    def __len__(self):
        return len(self.manager.source_instance)

    def __iter__(self):
        r"""Returns an iterator over the dataset with all transforms applied"""
        return self.manager.fast_compute()

    def assequence(self, of=None):
        r"""Returns a sequence of `{key: computed_value for key in of}`

        Args:
            of (str|set(str), optional): the keywords to keep in each element.
                Defaults to None.
        """
        if of is None:
            return iter(self)
        return map(
            lambda data: {key: value for key, value in data.items() if key in of}, self
        )

    def set_arguments_for(self, cls, *args, **kwargs):
        r"""Provide arguments to use when creating cls instances.
        Dame will use cls(*args, **kwargs) to get an instance.

        Args:
            cls (transform | source class): Class for which params will be registered
            *args (any): list of positional args
            **kwargs (any): list of keyword args
        """
        ctx = {"args": args, "kwargs": kwargs}
        self.context[cls.__name__] = ctx
        return self

    def add_arguments_for(self, cls, *args, **kwargs):
        if cls.__name__ not in self.context:
            self.context[cls.__name__] = {}
        ctx = self.context[cls.__name__]
        ctx["args"] = ctx.get("args", []) + args
        if "kwargs" not in ctx:
            ctx["kwargs"] = {}
        ctx["kwargs"].update(kwargs)

        return self

    def validate(self):
        r"""Validates (roughly) the subclass's transforms and source

        This method is called each time a subclass of dataset is created.
        It ensures that the transforms have 'provides' attribute and
        that no provided keywords overlap.

        Args:
            self (Dataset): The new instance of Dataset
        """
        assert (
            getattr(self, "source", None) is not None
        ), "Dataset can't exist without a source. "
        repeated_keywords = [
            kw for kw, num in Counter(self._get_all_keywords()).most_common() if num > 1
        ]
        assert len(repeated_keywords) == 0, (
            "Transforms' provided keywords must be unique. "
            f"Violated by {repeated_keywords}"
        )
        return self

    def _get_all_keywords(self):
        r"""Gets all the keywords declared by a dataset.

        Args:
            obj (Dataset or cls): The dataset object or a subclass of a Dataset

        Returns:
            list(string): all keywords declared by obj.
        """

        keywords = [
            chain.from_iterable(
                [t.provides for t in chain(self.transforms, (self.source,))]
            )
        ]
        return keywords
